/*		Butterworth filters coded by Paris Smaragdis 1994	*/
/*		Berklee College of Music Csound development team	*/
/*		Copyright (c) May 1994.  All rights reserved		*/

#include "cs.h"
#include "butter.h"
#include <math.h>

void butter_filter(float *, float *, float *);

void butset(BFIL *p)			/*	Hi/Lo pass set-up	*/
{
	p->a[6] = p->a[7] = 0.0;
	p->lkf = 0.0;
}

void bbutset(BBFIL *p)			/*	Band set-up	*/
{
	p->a[6] = p->a[7] = 0.0;
	p->lkb = 0.0;
	p->lkf = 0.0;
}

void hibut(BFIL *p)			/*	Hipass filter	*/
{
	register float	*out, *in;

	in = p->ain;
	out = p->sr;

	if (*p->kfc <= 0.0)	{
		register long	n = ksmps;

		do	*out++ = *in++;
		while (--n);

		return;
	}

	if (*p->kfc != p->lkf)	{	
		register float	*a, c, root2;

		a = p->a;
		root2 = sqrt( 2.0);
		c = tan( pi * *p->kfc / esr);
		a[1] = 1.0 / ( 1.0 + root2 * c + c * c);
		a[2] = -2.0 * a[1];
		a[3] = a[1];
		a[4] = 2.0 * ( c*c - 1.0) * a[1];
		a[5] = ( 1.0 - root2 * c + c * c) * a[1];
		p->lkf = *p->kfc;
	} 

	butter_filter( in, out, p->a);
}

void lobut(BFIL *p)				/*	Lopass filter	*/
{
	register float	*out, *in;

	in = p->ain;
	out = p->sr;

	if (*p->kfc <= 0.0)	{
		register long	n = ksmps;

		do	*out++ = 0.0;
		while (--n);

		return;
	}

	if (*p->kfc != p->lkf)	{	
		register float	*a, c, root2;

		a = p->a;
		root2 = sqrt(2.0);
		c = 1.0 / tan( pi * *p->kfc / esr);
		a[1] = 1.0 / ( 1.0 + root2 * c + c * c);
		a[2] = 2.0 * a[1];
		a[3] = a[1];
		a[4] = 2.0 * ( 1.0 - c*c) * a[1];
		a[5] = ( 1.0 - root2 * c + c * c) * a[1];
		p->lkf = *p->kfc;
	}
	
	butter_filter( in, out, p->a);
}

void bpbut(BBFIL *p)				/*	Bandpass filter	*/
{
	register float	*out, *in;

	in = p->ain;
	out = p->sr;

	if (*p->kbw <= 0.0)	{
		register long	n = ksmps;

		do	*out++ = 0.0;
		while (--n);

		return;
	}

	if (*p->kbw != p->lkb || *p->kfo != p->lkf)	{
		register float *a, c, d;

		a = p->a;
		c = 1.0 / tan( pi * *p->kbw / esr);
		d = 2.0 * cos( 2.0 * pi * *p->kfo / esr);
		a[1] = 1.0 / ( 1.0 + c);
		a[2] = 0.0;
		a[3] = -a[1];
		a[4] = - c * d * a[1];
		a[5] = ( c - 1) * a[1];
		p->lkf = *p->kfo;
		p->lkb = *p->kbw;
	}

	butter_filter( in, out, p->a);
}

void bcbut(BBFIL *p)						/*	Band reject filter	*/
{
	register float	*out, *in;

	in = p->ain;
	out = p->sr;

	if (*p->kbw <= 0.0)	{
		register long	n = ksmps;

		do	*out++ = *in++;
		while (--n);

		return;
	}

	if (*p->kbw != p->lkb || *p->kfo != p->lkf)	{
		register float *a, c, d;

		a = p->a;
		c = tan( pi * *p->kbw / esr);
		d = 2.0 * cos( 2.0 * pi * *p->kfo / esr);
		a[1] = 1.0 / ( 1.0 + c);
		a[2] = - d * a[1];
		a[3] = a[1];
		a[4] = a[2];
		a[5] = ( 1.0 - c) * a[1];
		p->lkf = *p->kfo;
		p->lkb = *p->kbw;
	}

	butter_filter( in, out, p->a);
}

void butter_filter(float *in, float *out, float *a)		/*	Filter loop	*/
{
	register float t, y, n = ksmps;

	do {
		t = *in++ - a[4] * a[6] - a[5] * a[7];
		y = t * a[1] + a[2] * a[6] + a[3] * a[7];
		a[7] = a[6];
		a[6] = t;
		*out++ = y;
	} while (--n);
}
