#include "cs.h"                                             /*  LPANAL.C      */
#include "soundio.h"
#include "lpc.h"
#include "window.h"
#include <math.h>

/* LPC analysis, modified by BV 8'92 for linkage to audio files via soundin.c.
 * Currently set for maximum of 50 poles, & max anal segment of 1000 samples,
 * meaning that the frame slices cannot exceed 500 samples.
 * Program size expands linearly with slice size, and as square of npoles.
 */

#define DEFNPOLES 34        /* recommeded default (max 50 in lpc.h)    */
#define DEFSLICE 200        /* <= MAXWINDIN/2 (currently 1000 in lpc.h) */
#define PITCHMIN        70.0
#define PITCHMAX        200.0   /* default limits in Hz for pitch search */

static  int     NPOLES, WINDIN, debug=0, verbose=0, doPitch = 1;
static  double  *x;
static  double  (*a)[MAXPOLES];
static  void    alpol(float *, double *, double *, double *, double *);
static  void    gauss(double (*)[MAXPOLES], double*, double*);
static  void    quit(char *), lpdieu(char *), usage(void);
extern  void    ptable(float, float, float, int);
extern  float   getpch(float*);
extern  int     openout(char*, int);
extern  OPARMS  O;
#ifdef __ZPC__
extern  int     winfg_continuous;
#endif

#ifdef TRACE
FILE *trace;
#endif

static  WINDAT   pwindow;

#define FIND(MSG)   if (*s == '\0')  \
                        if (!(--argc) || ((s = *++argv)!=0) && *s == '-')  \
                            lpdieu(MSG);

int lpanal(int argc, char **argv)
{
        int     slice, analframes, counter, channel;
        float   coef[NDATA+MAXPOLES], beg_time, input_dur, sr = 0.;
        char    *infilnam, *outfilnam;
        int     infd, ofd;
        double  errn, rms1, rms2, cc[MAXPOLES];
        float   *sigbuf, *sigbuf2;      /* changed from short */
        long    n;
        int     nb, osiz, hsize;
        LPHEADER    *lph;
        char    *lpbuf, *tp;
        float   pchlow, pchhigh;
        SOUNDIN *p;             /* struct allocated by SAsndgetset */
extern  int     SAsndgetset(char*, SOUNDIN **, float*, float*, float*, int);
extern  long    getsndin(int, float*, long, SOUNDIN*);
extern  char    *retfilnam;

        lpbuf = mcalloc((long)LPBUFSIZ);
        lph = (LPHEADER *) lpbuf;
        tp = lph->text;

        NPOLES = DEFNPOLES;         /* DEFAULTS... */
        slice = DEFSLICE;
        channel = 1;
        beg_time = 0.0;
        input_dur = 0.0;                /* default duration is to length of ip */
        *tp = '\0';
        pchlow = PITCHMIN;
        pchhigh = PITCHMAX;
        O.displays = 0;

        if (!(--argc))
#ifdef CWIN
        {
extern int __cdecl lpanal_dialog(float *, int *, float *, float *, int *, int *,
                                 float *, float *, int *, int *, char **, char **);
            if (!lpanal_dialog(&sr, &channel, &beg_time, &input_dur, &NPOLES, &slice, &pchlow,
                               &pchhigh, &verbose, &O.displays, &infilnam, &outfilnam)) return 0;
            if (pchlow == 0.) doPitch = 0;      /* -P0 inhibits ptrack */
            if (verbose > 1) debug = 1;
        }
        else {
#else
            lpdieu("insufficient arguments");
#endif
        do {
            register char *s = *++argv;
            if (*s++ == '-')
                switch (*s++) {
                case 's':       FIND("no sampling rate")
                    sscanf(s,"%f",&sr); break;
                case 'c':       FIND("no channel")
                    sscanf(s,"%d",&channel); break;
                case 'b':       FIND("no begin time")
                    sscanf(s,"%f",&beg_time); break;
                case 'd':       FIND("no duration time")
                    sscanf(s,"%f",&input_dur); break;
                case 'p':       FIND("no poles")
                    sscanf(s,"%d",&NPOLES); break;
                case 'h':       FIND("no hopsize")
                    sscanf(s,"%d",&slice); break; 
                case 'C':       FIND("no comment string")
                    strncat(tp,s,(LPBUFSIZ - sizeof(LPHEADER) + 4));
                    tp += strlen(tp);
                    break;
                case 'P':       FIND("no low frequency")
                    sscanf(s,"%f",&pchlow);
                    if (pchlow == 0.) doPitch = 0;      /* -P0 inhibits ptrack */
                    break;
                case 'Q':       FIND("no high frequency")
                    sscanf(s,"%f",&pchhigh); break;
                case 'v':       FIND("no verbose level")
                    sscanf(s,"%d",&verbose);
                    if (verbose > 1)  debug = 1;
                    break;
                case 'g':       O.displays = 1;
#ifdef __ZPC__
                                winfg_continuous = 1;
#endif
                    break;
		case '-':
		    FIND("no log file");
		    dribble = fopen(s, "w");
		    while (*s++); s--;
		    break;
                default: sprintf(errmsg,"unrecognized flag -%c", *--s);
                         lpdieu(errmsg);
                }
            else break;
        } while (--argc);

        if (argc != 2)  lpdieu("incorrect number of filenames");
        infilnam = *argv++;
        outfilnam = *argv;
#ifdef CWIN
        }
#endif
        if (NPOLES > MAXPOLES)
            quit("poles exceeds maximum allowed");
        if (slice < NPOLES * 5)
            warning("hopsize may be too small, recommend at least npoles * 5");
        if ((WINDIN = slice * 2) > MAXWINDIN)
            quit("input framesize (inter-frame-offset*2) exceeds maximum allowed");

        if (verbose) {
            err_printf("Reading sound from %s, writing lpfile to %s\n",
                    infilnam, outfilnam);
            err_printf("poles=%d hopsize=%d begin=%4.1f duration=%4.1f\n",
                    NPOLES, slice, beg_time, input_dur);
            err_printf("lpheader comment:\n%s\n", lph->text);
            if (pchlow > 0.)
                err_printf("pch track range: %5.1f - %5.1f Hz\n",pchlow,pchhigh);
            else err_printf("pitch tracking inhibited\n");
        }
        if ((input_dur < 0) || (beg_time < 0))
            quit("input and begin times cannot be less than zero");
                
        if (!(infd = SAsndgetset(infilnam,&p,&beg_time,&input_dur,&sr,channel))) {
            sprintf(errmsg,"error while opening %s", retfilnam);
            quit(errmsg);
        }

        if ((ofd = openout(outfilnam, 1)) < 0)  /* open output file */
            quit("cannot create output file");
        lph->lpmagic = LP_MAGIC;                /* build output hdr */
        lph->npoles = NPOLES;
        lph->nvals = NPOLES + NDATA;
        lph->srate = p->sr;
        lph->framrate = (float) p->sr / slice;
        lph->duration = input_dur;
        hsize = tp - (char *) lph;              /* header size including text */
        lph->headersize = (hsize + 3) & -4;     /* rounded up to 4 byte bndry */
        if (lph->headersize > LPBUFSIZ)    /* UNNECESSARY ?? */
            lph->headersize = LPBUFSIZ;
        if ((nb = write(ofd,(char *)lph,(int)lph->headersize)) < lph->headersize)
            quit("can't write header");
        osiz = (NPOLES + NDATA) * sizeof(float);

        sigbuf = (float *) mmalloc((long)WINDIN * sizeof(float));
        sigbuf2 = sigbuf + slice;
        if ((n = getsndin(infd, sigbuf, (long)WINDIN, p)) < WINDIN)
            quit("soundfile read error, couldn't fill first frame");
        if (doPitch)
            ptable(pchlow, pchhigh, (float) p->sr, WINDIN);
        counter = 0;
        analframes = (p->getframes - 1) / slice;
        dispinit();
        dispset(&pwindow,coef+4,NPOLES,"pitch: 0000.00   ",0,"LPC/POLES");
        a = (double (*)[MAXPOLES]) malloc(NPOLES * MAXPOLES * sizeof(double));  /* Space for a array */
        x = (double *) malloc(WINDIN * sizeof(double));  /* alloc a double array */
#ifdef TRACE
	trace = fopen("lpanal.trace", "w");
#endif
        do {
            register float *fp1;
            register double *dfp;

            counter++;
            alpol(sigbuf, &errn, &rms1, &rms2, cc);
            coef[0] = (float)rms2; 
            coef[1] = (float)rms1; 
            coef[2] = (float)errn; 
            if (doPitch)
                coef[3] = getpch(sigbuf);
            else coef[3] = 0.0;
            if (debug) err_printf("%d\t%9.4f\t%9.4f\t%9.4f\t%9.4f\n",
                               counter, coef[0], coef[1], coef[2], coef[3]);
#ifdef TRACE
	    if (debug) fprintf(trace,"%d\t%9.4f\t%9.4f\t%9.4f\t%9.4f\n",
			       counter, coef[0], coef[1], coef[2], coef[3]);
#endif
            for (fp1=coef+NDATA, dfp=cc+NPOLES, n=NPOLES; n--; )
                *fp1++ = - (float) *--dfp;               /* rev coefs & chng sgn */
            sprintf(pwindow.caption, "pitch: %8.2f", coef[3]);
            display(&pwindow);
            if ((nb = write(ofd, (char *)coef, osiz)) != osiz)  /* wrt anal fram */
                quit("write error");
            memcpy(sigbuf, sigbuf2, sizeof(float)*slice);
/*          ( float *fp2; for (fp1=sigbuf, fp2=sigbuf2, n=slice; n--; ) */ /* move slice forward */
/*              *fp1++ = *fp2++;} */
            if ((n = getsndin(infd, sigbuf2, (long)slice, p)) == 0)
                break;                                          /* refil til EOF */
            POLL_EVENTS();
        } while (counter < analframes);                         /* or nsmps done */
        dispexit();
        printf("%d lpc frames written to %s\n", counter, outfilnam);
        close(infd);
        close(ofd);
        free(a); free(x);
        exit(0);
}

static void quit(char *msg)
{
        printf("lpanal: %s\n", msg);
        die("analysis aborted");
}

static void lpdieu(char *msg)
{
        printf("lpanal: %s\n", msg);
        usage();
        exit(0);
}

static void alpol(float *sig, double *errn, double *rms1, double *rms2, double b[MAXPOLES])
                                        /* sig now float */
                                        /* b filled here */
{
        double v[MAXPOLES];
        double *xp;
        double sum, sumx, sumy;
        int i, j, k, limit;

        for (xp=x; xp-x < WINDIN ;++xp,++sig)            /*   &  copy sigin into */
            *xp = (double) *sig;
        for (i=0; i < NPOLES ;++i)  {
            sum = (double) 0.0;
            for (k=NPOLES; k < WINDIN ;++k)
                sum += x[k-(i+1)] * x[k];
            v[i] = -sum;
            if (i != NPOLES - 1)  {
                limit = NPOLES - (i+1);
                for (j=0; j < limit; j++)  {
                    sum += x[NPOLES-(i+1)-(j+1)] * x[NPOLES-(j+1)]
                         - x[WINDIN-(i+1)-(j+1)] * x[WINDIN-(j+1)];
                    a[(i+1)+j][j] = a[j][(i+1)+j] = sum;
                }
            }
        }
        sum = (double) 0.0;
        for (k=NPOLES; k < WINDIN ;++k)
            sum +=  x[k]*x[k];
        sumy = sumx = sum;
#ifdef TRACE
	fprintf(trace, "sumy=%lf\n", sumy);
#endif
        for (j=0; j < NPOLES; j++)  {
            sum += x[NPOLES-(j+1)]*x[NPOLES-(j+1)] - 
                   x[WINDIN-(j+1)]*x[WINDIN-(j+1)];
            a[j][j] = sum;
        }
        gauss(a, v, b);
/*      filtn(x, y, b);   */
        for (i=0; i < NPOLES ;++i) {
            sumy -= b[i]*v[i];
#ifdef TRACE
	    fprintf(trace, "b[i]=%lf v[i]=%lf sumy=%lf\n", b[i], v[i], sumy);
#endif
        }
        *rms1 = sqrt(sumx / (WINDIN - NPOLES) );
        *rms2 = sqrt(sumy / (WINDIN - NPOLES) );
/*        sum = (*rms2)/(*rms1);  *errn = sum*sum; */
	*errn = sumy/sumx;
}

static void gauss(double (*a/*old*/)[MAXPOLES], double *bold, double b[])
{
        double amax, dum, pivot;
        double c[MAXPOLES];
        int i, j, k, l, istar, ii, lp;

        /* bold untouched by this subroutine */
        for (i=0; i < NPOLES ;++i)  {
            c[i] = bold[i];
#ifdef TRACE
	    fprintf(trace, "c[%d]=%lf\n", i, c[i]);
#endif
        }
        /* eliminate i-th unknown */
        for (i=0; i < NPOLES - 1 ;++i)  {        /* find largest pivot */
            amax = 0.0;
            for (ii=i; ii < NPOLES ;++ii)  {
                if (fabs(a[ii][i]) >= amax)  {
                    istar = ii;
                    amax = fabs(a[ii][i]);
                }
            }
            if (amax < 1e-20)
                die("gauss: ill-conditioned");
	    if (i != istar) {
	      for (j=0; j < NPOLES ;++j)  {    /* switch rows */
                dum = a[istar][j];
                a[istar][j] = a[i][j];
                a[i][j] = dum;
	      }
	      dum = c[istar];
	      c[istar] = c[i];
	      c[i] = dum;
	    }
#ifdef TRACE
	    fprintf(trace, "c[%d]=%lf\n", i, c[i]);
#endif
            for (j=i+1; j < NPOLES ;++j)  {             /* pivot */
                pivot = a[j][i] / a[i][i];
#ifdef TRACE
		fprintf(trace, "pivot = %lf c[%d] = %lf\n", pivot, j, c[j]);
#endif
                c[j] = c[j] - pivot * c[i];
#ifdef TRACE
		fprintf(trace, "c[%d]=%lf c[%d]=%lf\n", j, c[j], i, c[i]);
#endif
                for (k=0; k < NPOLES ;++k)
                    a[j][k] = a[j][k] - pivot * a[i][k];
            }
        }                               /* return if last pivot is too small */
        if (fabs(a[NPOLES-1][NPOLES-1]) < 1e-20)
            die("gauss: ill-conditioned");

        b[NPOLES-1] = c[NPOLES-1] / a[NPOLES-1][NPOLES-1];
#ifdef TRACE
	fprintf(trace, "b[NPOLES-1]=%lf c[NPOLES-1]=%lf a[][]=%lf\n",
		b[NPOLES-1], c[NPOLES-1], a[NPOLES-1][NPOLES-1]);
#endif
        for (k=0; k<NPOLES-1; ++k)  {   /* back substitute */
            l = NPOLES-1 -(k+1);
            b[l] = c[l];
            lp = l + 1;
            for (j = lp; j<NPOLES; ++j) {
                b[l] -= a[l][j] * b[j];
            }
            b[l] /= a[l][l];
        }
}

static void usage()
{
  err_printf("USAGE:\tlpanal [flags] infilename outfilename\n");
  err_printf("\twhere flag options are:\n");
  err_printf("-s<srate>\tinput sample rate (defaults to header else %7.1f.)\n",
          DFLT_SR);
  err_printf("-c<chnlreq>\trequested channel of sound (default chan 1)\n");
  err_printf("-b<begin>\tbegin time in seconds into soundfile (default 0.0)\n");
  err_printf("-d<duration>\tseconds of sound to be analyzed (default: to EOF)\n");
  err_printf("-p<npoles>\tnumber of poles for analysis (default %d)\n",
          DEFNPOLES);
  err_printf("-h<hopsize>\toffset between frames in samples (default %d)\n",
          DEFSLICE);
  err_printf("\t\t\t(framesize will be twice <hopsize>)\n");
  err_printf("-C<string>\tcomment field of lp header (default empty)\n");
  err_printf("-P<mincps>\tlower limit for pitch search (default %5.1f Hz)\n",
          PITCHMIN);
  err_printf("\t\t\t(-P0 inhibits pitch tracking)\n");
  err_printf("-Q<maxcps>\tupper limit for pitch search (default %5.1f Hz)\n",
          PITCHMAX);
  err_printf("-v<verblevel>\tprinting verbosity: 0=none, 1=verbose, 2=debug.");
  err_printf(" (default 0)\n");
  err_printf("-g\tgraphical display of results\n");
  err_printf("-- fname\tLog output to file\n");
  err_printf("see also:  Csound Manual Appendix\n");
}
