#include "cs.h"                                       /*    MIDIOPS.C    */
#include "midiops.h"
#include <math.h>

static float f128 = 128.;
extern float *cpsocint, *cpsocfrc, octresol;
extern INSDS    *curip;         /* valid at I-time */

void notnum(MIDIVAL *p)      /* valid only at I-time */
{
    *p->r = curip->m_pitch;
}

void veloc(MIDIVAL *p)          /* valid only at I-time */
{
    *p->r = curip->m_veloc;
}

void pchmidi(MIDIVAL *p)
{
    register INSDS *curip = p->h.insdshead;
    double fract, oct, ioct;
    oct = curip->m_pitch / 12. + 3.;
    fract = modf(oct, &ioct);
    fract *= 0.12;
    *p->r = ioct + fract;
}

void pchmidib(MIDIVAL *p)
{
    register INSDS *curip = p->h.insdshead;
    double fract, oct, ioct;
    oct = (curip->m_pitch + curip->m_chnbp->pbendiff) / 12. + 3.;
    fract = modf(oct, &ioct);
    fract *= 0.12;
    *p->r = ioct + fract;
}

void octmidi(MIDIVAL *p)
{
    register INSDS *curip = p->h.insdshead;
    *p->r = curip->m_pitch / 12. + 3.;
}

void octmidib(MIDIVAL *p)
{
    register INSDS *curip = p->h.insdshead;
    *p->r = (curip->m_pitch + curip->m_chnbp->pbendiff) / 12. + 3.;
}

void cpsmidi(MIDIVAL *p)
{
    register INSDS *curip = p->h.insdshead;
    register long  loct;
    loct = (curip->m_pitch / 12. + 3.) * octresol;
    *p->r = CPSOCTL(loct);
}

void cpsmidib(MIDIVAL *p)
{
    register INSDS *curip = p->h.insdshead;
    register long  loct;
    loct = ((curip->m_pitch + curip->m_chnbp->pbendiff) / 12. + 3.) * octresol;
    *p->r = CPSOCTL(loct);
}

void ampmidi(MIDIAMP *p)	/* convert midi veloc to amplitude */
				/*   valid only at I-time          */
{
    float amp;
    long  fno;
    register FUNC *ftp;

    amp = curip->m_veloc / f128;                    /* amp = normalized veloc */
    if ((fno = *p->ifn) > 0) {
	if ((ftp = ftfind(p->ifn)) == NULL)         /* if valid ftable,       */
	    return;                                 /*     use amp as index   */
	amp = *(ftp->ftable + (long)(amp * ftp->flen));
    }
    *p->r = amp * *p->imax;                         /* now scale the output   */
}

void aftset(MIDISCL *p)
{
    p->scale = *p->iscal / f128;
}

void aftouch(MIDISCL *p)
{
    register INSDS *curip = p->h.insdshead;
    *p->r = curip->m_chnbp->katouch[curip->m_pitch] * p->scale;
}

void chpress(MIDISCL *p)
{
    register INSDS *curip = p->h.insdshead;
    *p->r = curip->m_chnbp->chnpress * p->scale;
}

void pbenset(MIDISCL *p)
{
    p->scale = *p->iscal;
}

void pchbend(MIDISCL *p)
{
    register INSDS *curip = p->h.insdshead;
    *p->r = curip->m_chnbp->pbendiff * p->scale;
}

void imidictl(MIDICTL *p)
{
    register long  ctlno;
    if ((ctlno = *p->ictlno) < 0 || ctlno > 120)
	initerror("illegal controller number");
    else *p->r = curip->m_chnbp->ctl_val[ctlno];
}

void mctlset(MIDICTL *p)
{
    register long  ctlno;
    if ((ctlno = *p->ictlno) < 0 || ctlno > 120)
	initerror("illegal controller number");
    else p->ctlno = ctlno;
}

void midictl(MIDICTL *p)
{
    register INSDS *curip = p->h.insdshead;
    *p->r = curip->m_chnbp->ctl_val[p->ctlno];
}

void midiolap(MIDIOLAP *p)	/* set overlap factor for MIDI voices */
				/*   valid only at I-time          */
{
    curip->overlap = (short) *p->olap - 1;
}

void midiage(MIDIAGE *p)	/* during performance, reads current age */
				/* of voice */            
{
    register INSDS *curip = p->h.insdshead;
    *p->r = (float) curip->age;                                         
}    
